from ctypes import Array as ctypes_Array
from _ctypes import _CData
from collections.abc import Mapping, MutableMapping
from numbers import Real
from selectors import _PollLikeSelector
from threading import Lock
from typing import ClassVar, List, NamedTuple, Optional, Protocol, Tuple, TypeVar, Union

from ._util.wintypes_extra import WSAPOLLFD

T = TypeVar('T')

class _SelectorKey(NamedTuple):
    fileobj: Union[_Fileobj, int]
    fd: int
    events: int
    data: T

POLLERR: int
POLLHUP: int
POLLIN: int
POLLNVAL: int
POLLOUT: int
POLLPRI: int
POLLRDBAND: int
POLLRDNORM: int
POLLWRBAND: int
POLLWRNORM: int

class _Fileobj(Protocol):
    def fileno(self) -> int: ...

class wsapoll:
    __slots__ = ['__impl', '__fd_to_key', '__buffer']
    _registered: MutableMapping[int, int]
    __impl: ctypes_Array[WSAPOLLFD]
    __impl_uptodate: bool
    __buffer: _CData
    __lock: Lock
    def __init__(self, sizehint: int=...) -> None: ...
    def __repr__(self) -> str: ...

    def register(self, fd: Union[_Fileobj, int], eventmask: int=...) -> None: ...
    def modify(self, fd: Union[_Fileobj, int], eventmask: int) -> None: ...
    def unregister(self, fd: Union[_Fileobj, int]) -> None: ...
    def poll(self, timeout: Optional[Real]=None) -> List[Tuple[int, int]]: ...

    def _poll(self, timeout: int=-1) -> List[Tuple[int, int]]: ...
    def _clear(self) -> None: ...
    def __check_maybe_affected(self) -> bool: ...
    def _selectors_close_impl(self) -> None: ...
    def __update_impl(self) -> None: ...
    def __getstate__(self) -> Mapping[int, int]: ...
    def __setstate__(self, state: Mapping[int, int]) -> None: ...

class WSAPollSelector(_PollLikeSelector):
    def register(self, fileobj: Union[_Fileobj, int], events: int, data: T=None) -> _SelectorKey: ...
    def unregister(self, fileobj: Union[_Fileobj, int]) -> _SelectorKey: ...
    def modify(self, fileobj: Union[_Fileobj, int], events: int, data: T=None) -> _SelectorKey: ...
    def select(self, timeout: Optional[Real]=None) -> List[Tuple[_SelectorKey, int]]: ...
    def close(self) -> None: ...
    def get_key(self, fileobj: Union[_Fileobj, int]) -> _SelectorKey: ...
    def get_map(self) -> Mapping[Union[_Fileobj, int], _SelectorKey]: ...

    _selector_cls: ClassVar[...]  # type: ignore
    _EVENT_READ: ClassVar[int]
    _EVENT_WRITE: ClassVar[int]



import winpoll

import contextlib
import time
import socket
import unittest

UNDERSLEEP_TOLERANCE_NS = 0  # zero-tolerance
OVERSLEEP_TOLERANCE_NS = 50_000_000  # 50 milliseconds


def _churn_fds(minimum=1024):
    deadline_ns = time.monotonic_ns() + 10_000_000_000  # 10 seconds
    while True:
        with contextlib.ExitStack() as stack:
            socks = [stack.enter_context(sock) for sock in socket.socketpair()]

            if any(sock.fileno() > minimum for sock in socks):
                return

            if time.monotonic_ns() > deadline_ns:
                raise RuntimeError(f"failed to reach target minimum fd")


class TestWsaPoll(unittest.TestCase):
    def test_pollout(self):
        obj = winpoll.wsapoll()

        with contextlib.ExitStack() as stack:
            sock1, sock2 = map(stack.enter_context, socket.socketpair())
            obj.register(sock1, winpoll.POLLOUT)

            result = obj.poll()

            self.assertEqual(len(result), 1)
            # !!!!!FIXME!!!!! don't run the following code if the preceding test failed
            result_fd, result_events = result[0]
            self.assertEqual(result_fd, sock1.fileno())
            self.assertEqual(result_events, winpoll.POLLWRNORM)

    def test_pollin(self):
        obj = winpoll.wsapoll()

        with contextlib.ExitStack() as stack:
            sock1, sock2 = map(stack.enter_context, socket.socketpair())
            obj.register(sock2, winpoll.POLLIN)

            sock1.send(b"*")
            result = obj.poll()

            self.assertEqual(len(result), 1)

            result_fd, result_events = result[0]
            self.assertEqual(result_fd, sock2.fileno())
            self.assertEqual(result_events, winpoll.POLLRDNORM)

    def test_timeout(self):
        obj = winpoll.wsapoll()

        with contextlib.ExitStack() as stack:
            sock1, sock2 = map(stack.enter_context, socket.socketpair())
            obj.register(sock2, winpoll.POLLIN)
            obj.poll(0)  # ensure the pollfd buffer is up-to-date

            for timeout in range(5):
                timeout_ns = timeout * 1_000_000_000

                with self.subTest(timeout=timeout):
                    t0 = time.monotonic_ns()
                    result = obj.poll(timeout)
                    t1 = time.monotonic_ns()

                    self.assertEqual(len(result), 0)

                    dt = t1 - t0
                    self.assertGreaterEqual(dt, timeout_ns - UNDERSLEEP_TOLERANCE_NS)
                    self.assertLess(dt, timeout_ns + OVERSLEEP_TOLERANCE_NS)
